from PyQt5.QtWidgets import QDialog, QLabel, QComboBox, QPushButton, QApplication
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QStandardItemModel, QStandardItem


class DeleteUserDialog(QDialog):
    """
    Класс, определяющий и создающий окно удаления пользователя
    """

    def __init__(self, db, server):
        super().__init__()
        self.db = db
        self.server = server

        self.setFixedSize(350, 120)
        self.setWindowTitle('Удаление пользователя')
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(True)

        self.selector_label = QLabel(
            'Выберите пользователя для удаления:', self)
        self.selector_label.setFixedSize(200, 20)
        self.selector_label.move(10, 0)

        self.selector = QComboBox(self)
        self.selector.setFixedSize(200, 20)
        self.selector.move(10, 30)

        self.ok_button = QPushButton('Удалить', self)
        self.ok_button.setFixedSize(100, 30)
        self.ok_button.move(230, 20)
        self.ok_button.clicked.connect(self.remove_user)

        self.cancel_button = QPushButton('Отмена', self)
        self.cancel_button.setFixedSize(100, 30)
        self.cancel_button.move(230, 60)
        self.cancel_button.clicked.connect(self.close)

        self.all_users_fill()

    def all_users_fill(self):
        """
        Метод, заполняющий меню выбора пользователя пользователями из базы данных
        """

        self.selector.addItems([item[0] for item in self.db.get_users()])

    def remove_user(self):
        """
        Метод, провоцирующий удаление пользователя из базы данных сервера и отключающий его от сервера
        """

        self.db.delete_user(self.selector.currentText())
        if self.selector.currentText() in self.server.names:
            sock = self.server.names[self.selector.currentText()]
            del self.server.names[self.selector.currentText()]
            self.server.remove_client(sock)
        self.server.service_update_lists()
        self.close()


if __name__ == '__main__':
    app = QApplication([])
    from server_db import ServerDB
    db = ServerDB('../../../server_db.db3')
    import os
    import sys
    path1 = os.path.join(os.getcwd(), '../../..')
    sys.path.insert(0, path1)
    from core import MessageProcessor
    server = MessageProcessor('127.0.0.1', 7777, db)
    dial = DeleteUserDialog(db, server)
    dial.show()
    app.exec_()



from PyQt5.QtWidgets import QDialog, QPushButton, QLineEdit, QApplication, QLabel, QMessageBox
from PyQt5.QtCore import Qt
import hashlib
import binascii


class AddUserDialog(QDialog):
    """
    Класс, определяющий и создающий окно создания нового пользователя
    """

    def __init__(self, db, server):
        super().__init__()

        self.db = db
        self.server = server

        self.setWindowTitle('Регистрация')
        self.setFixedSize(175, 183)
        self.setModal(True)
        self.setAttribute(Qt.WA_DeleteOnClose)

        self.label_username = QLabel('Введите имя пользователя:', self)
        self.label_username.move(10, 10)
        self.label_username.setFixedSize(150, 15)

        self.client_name = QLineEdit(self)
        self.client_name.setFixedSize(154, 20)
        self.client_name.move(10, 30)

        self.label_passwd = QLabel('Введите пароль:', self)
        self.label_passwd.move(10, 55)
        self.label_passwd.setFixedSize(150, 15)

        self.client_passwd = QLineEdit(self)
        self.client_passwd.setFixedSize(154, 20)
        self.client_passwd.move(10, 75)
        self.client_passwd.setEchoMode(QLineEdit.Password)
        self.label_conf = QLabel('Введите подтверждение:', self)
        self.label_conf.move(10, 100)
        self.label_conf.setFixedSize(150, 15)

        self.client_conf = QLineEdit(self)
        self.client_conf.setFixedSize(154, 20)
        self.client_conf.move(10, 120)
        self.client_conf.setEchoMode(QLineEdit.Password)

        self.ok_button = QPushButton('Сохранить', self)
        self.ok_button.move(10, 150)
        self.ok_button.clicked.connect(self.save_data)

        self.cancel_button = QPushButton('Выход', self)
        self.cancel_button.move(90, 150)
        self.cancel_button.clicked.connect(self.close)

        self.messages = QMessageBox()

        self.show()

    def save_data(self):
        """
        Метод, проверяющий введённые данные, хэширующий введённый пароль и записывающий нового пользователя в базу
        данных сервера
        """

        if not self.client_name.text():
            self.messages.critical(
                self, 'Ошибка', 'Не указано имя пользователя')
            return
        elif self.client_passwd.text() != self.client_conf.text():
            self.messages.critical(
                self, 'Ошибка', 'Введённые пароли не совпадают')
            return
        elif self.db.check_user(self.client_name.text()):
            self.messages.critical(
                self, 'Ошибка', 'Пользователь уже существует')
            return
        else:
            password_bytes = self.client_passwd.text().encode('utf-8')
            salt = self.client_name.text().lower().encode('utf-8')
            hash = hashlib.pbkdf2_hmac('sha512', password_bytes, salt, 10000)
            self.db.add_user(
                self.client_name.text(),
                binascii.hexlify(hash))
            self.messages.information(
                self, 'Успех', 'Пользователь успешно зарегистрирован.')
            self.server.service_update_lists()
            self.close()


if __name__ == '__main__':
    app = QApplication([])
    from server_db import ServerDB
    db = ServerDB('../../../server_db.db3')
    import os
    import sys
    path1 = os.path.join(os.getcwd(), '../../..')
    sys.path.insert(0, path1)
    from core import MessageProcessor
    server = MessageProcessor('127.0.0.1', 7777, db)
    dial = AddUserDialog(db, server)
    app.exec_()



from PyQt5.QtWidgets import QDialog, QLabel, QLineEdit, QPushButton, QFileDialog, QMessageBox
from PyQt5.QtCore import Qt
import os


class ConfigWindow(QDialog):
    """
    Класс, определяющий и создающий окно изменения настроек сервера
    """

    def __init__(self, config):
        super().__init__()
        self.config = config

        self.setWindowTitle('Настройки сервера')
        self.setFixedSize(365, 260)

        self.db_path_label = QLabel('Путь до файла базы данных: ', self)
        self.db_path_label.setFixedSize(250, 15)
        self.db_path_label.move(10, 10)

        self.db_path = QLineEdit(self)
        self.db_path.setFixedSize(250, 20)
        self.db_path.move(10, 30)
        self.db_path.setReadOnly(True)

        self.db_path_select = QPushButton('Обзор', self)
        self.db_path_select.move(275, 28)
        self.db_path_select.clicked.connect(self.open_file_dialog)

        self.db_file_label = QLabel('Имя файла базы данных: ', self)
        self.db_file_label.setFixedSize(180, 15)
        self.db_file_label.move(10, 68)

        self.db_file = QLineEdit(self)
        self.db_file.setFixedSize(150, 20)
        self.db_file.move(200, 66)

        self.port_label = QLabel('Номер порта для соединений: ', self)
        self.port_label.setFixedSize(180, 15)
        self.port_label.move(10, 108)

        self.port = QLineEdit(self)
        self.port.setFixedSize(150, 20)
        self.port.move(200, 108)

        self.ip_label = QLabel('С какого IP принимать соединения: ', self)
        self.ip_label.setFixedSize(180, 15)
        self.ip_label.move(10, 148)

        self.ip_label_note = QLabel('Оставьте это поле пустым, чтобы\n принимать соединения с любых адресов', self)
        self.ip_label_note.setFixedSize(500, 30)
        self.ip_label_note.move(10, 168)

        self.ip = QLineEdit(self)
        self.ip.setFixedSize(150, 20)
        self.ip.move(200, 148)

        self.save_button = QPushButton('Сохранить', self)
        self.save_button.move(190, 200)
        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(275, 220)
        self.close_button.clicked.connect(self.close)

        self.show()

        self.db_path.insert(self.config['SETTINGS']['Database_path'])
        self.db_file.insert(self.config['SETTINGS']['Database_file'])
        self.port.insert(self.config['SETTINGS']['Default_port'])
        self.ip.insert(self.config['SETTINGS']['Listen_Address'])
        self.save_button.clicked.connect(self.save_server_config)

    def open_file_dialog(self):
        """
        Метод, окрывающий окно определения пути до файла базы данных сервера
        """

        global dialog
        dialog = QFileDialog(self)
        path = dialog.getExistingDirectory()
        path = path.replace('/', '\\')
        self.db_path.clear()
        self.db_path.insert(path)

    def save_server_config(self):
        """
        Метод, проверяющий введённые данные на корректность и сохраняющий настройки в файле, если они корректны
        """

        global config_window
        message = QMessageBox()
        self.config['SETTINGS']['Database_path'] = self.db_path.text()
        self.config['SETTINGS']['Database_file'] = self.db_file.text()
        try:
            port = int(self.port.text())
        except ValueError:
            message.warning(self, 'Ошибка', 'Порт должен быть числом')
        else:
            self.config['SETTINGS']['Listen_Address'] = self.ip.text()
            if 1023 < port < 65536:
                self.config['SETTINGS']['Default_port'] = str(port)
                dir_path = os.path.dirname(os.path.realpath(__file__))
                dir_path = os.path.join(dir_path, '../../..')
                with open(f"{dir_path}/{'server_dist+++.ini'}", 'w') as conf:
                    self.config.write(conf)
                    message.information(
                        self, 'OK', 'Настройки успешно сохранены!')
            else:
                message.warning(
                    self, 'Ошибка', 'Порт должен быть от 1024 до 65536')



import threading
import select
import socket
import json
import hmac
import binascii
import os
import sys
sys.path.append('../../../')
from server_dist.server.common import send_message, get_message
from server_dist.server.common.decos import login_required

logger = logging.getLogger('server_dist')


class PortDescriptor:
    """
    Дескриптор, проверяющий, чтобы значение введённого порта было от 1024 до 65535
    """

    def __set_name__(self, owner, name):
        self.name = name

    def __set__(self, instance, value):
        if not 1023 < value < 65536:
            logger.critical(
                f'Попытка запуска сервера с указанием неподходящего порта {value}. Допустимы адреса с 1024 до 65535.')
            exit(1)
        instance.__dict__[self.name] = value


class MessageProcessor(threading.Thread):
    """
    Класс, реализующий функционал сервера
    """

    port = PortDescriptor()

    def __init__(self, address, port, db):
        super().__init__()
        self.daemon = True
        self.address = address
        self.port = port
        self.db = db
        self.transport = None

        self.clients = []
        self.messages = []
        self.names = {}
        self.receivers = None
        self.errors = None

        self.running = True

    def run(self):
        """
        Метод, запускающий сервер и реализующий функционал приёма и передачи сообщений
        """

        logger.info(
            f'Запущен сервер, порт для подключений: {self.port} , '
            f'адрес с которого принимаются подключения: {self.address}. '
            f'Если адрес не указан, принимаются соединения с любых адресов.')
        self.transport = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.transport.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.transport.bind((self.address, self.port))
        self.transport.settimeout(0.5)

        self.transport.listen()
        while self.running:
            try:
                client, client_address = self.transport.accept()
            except OSError:
                pass
            else:
                logger.info(f'Установлено соедение с ПК {client_address}')
                client.settimeout(5)
                self.clients.append(client)

            senders = []

            try:
                if self.clients:
                    senders, self.receivers, self.errors = select.select(self.clients, self.clients, [], 0)
            except OSError as e:
                logger.error(f'Ошибка работы с сокетами: {e}')

            if senders:
                for client_with_message in senders:
                    try:
                        self.process_client_message(get_message(client_with_message), client_with_message)
                    except (OSError, json.JSONDecodeError, TypeError) as err:
                        logger.debug(f'Getting data from client exception.', exc_info=err)
                        self.delete_client(client_with_message)

    def delete_client(self, client):
        """
        :param client: сокет клиента

        Метод, отключающий клиента от сервера и удаляющий его из словаря текущих клиентов
        """

        logger.info(f'Клиент {client.getpeername()} отключился от сервера.')
        for name in self.names:
            if self.names[name] == client:
                self.db.user_logout(name)
                del self.names[name]
                break
        self.clients.remove(client)
        client.close()

    def process_message(self, message):
        """
        :param message: словарь, содержащий данные об обрабатываемом сообщении

        Метод, обеспечивающий отправление сообщения получателю, если он зарегистрирован на сервере
        """

        if message[RECEIVER] in self.names and self.names[message[RECEIVER]] in self.receivers:
            try:
                send_message(self.names[message[RECEIVER]], message)
                logger.info(f'Отправлено сообщение пользователю {message[RECEIVER]} от пользователя {message[SENDER]}.')
            except OSError:
                self.delete_client(message[RECEIVER])
        elif message[RECEIVER] in self.names and self.names[message[RECEIVER]] not in self.receivers:
            logger.error(f'Связь с клиентом {message[RECEIVER]} была потеряна. Соединение закрыто, доставка невозможна.')
            self.delete_client(self.names[message[RECEIVER]])
        else:
            logger.error(
                f'Пользователь {message[RECEIVER]} не зарегистрирован на сервере, отправка сообщения невозможна.')

    @login_required
    def process_client_message(self, message, client):
        """
        :param message: словарь, содержащий данные об обрабатываемом сообщении
        :param client:  сокет клиента

        Метод, обрабатывающий сообщения  с командами от пользователя:
            -приветственное сообщение
            -сообщения другим пользователям
            -выход из программы
            -получение списка контактов пользователя
            -добавление контакта пользователя
            -удаление контакта пользователя
            -получение списка доступных пользователей
            -получение публичного ключа пользователя
        """

        logger.debug(f'Разбор сообщения от клиента : {message}')
        if ACTION in message and message[ACTION] == PRESENCE and TIME in message and USER in message:
            self.autorize_user(message, client)

        elif ACTION in message and message[ACTION] == MESSAGE and RECEIVER in message and TIME in message \
                and SENDER in message and MESSAGE_TEXT in message and self.names[message[SENDER]] == client:
            if message[RECEIVER] in self.names:
                self.db.process_message(
                    message[SENDER], message[RECEIVER])
                self.process_message(message)
                try:
                    send_message(client, RESPONSE_200)
                except OSError:
                    self.delete_client(client)
            else:
                response = RESPONSE_400
                response[ERROR] = 'Пользователь не зарегистрирован на сервере.'
                try:
                    send_message(client, response)
                except OSError:
                    pass
            return
        elif ACTION in message and message[ACTION] == EXIT and ACCOUNT_NAME in message \
                and self.names[message[ACCOUNT_NAME]] == client:
            self.delete_client(client)
        elif ACTION in message and message[ACTION] == GET_CONTACTS and ACCOUNT_NAME in message and \
                self.names[message[ACCOUNT_NAME]] == client:
            response = RESPONSE_202
            response[DATA] = self.db.get_contacts(message[ACCOUNT_NAME])
            try:
                send_message(client, response)
            except OSError:
                self.delete_client(client)
        elif ACTION in message and message[ACTION] == ADD_CONTACT and ACCOUNT_NAME in message and CONTACT in message \
                and self.names[message[ACCOUNT_NAME]] == client:
            self.db.add_contact(message[ACCOUNT_NAME], message[CONTACT])
            try:
                send_message(client, RESPONSE_200)
            except OSError:
                self.delete_client(client)
        elif ACTION in message and message[ACTION] == DEL_CONTACT and ACCOUNT_NAME in message and CONTACT in message \
                and self.names[message[ACCOUNT_NAME]] == client:
            self.db.delete_contact(message[ACCOUNT_NAME], message[CONTACT])
            try:
                send_message(client, RESPONSE_200)
            except OSError:
                self.delete_client(client)
        elif ACTION in message and message[ACTION] == GET_USERS and ACCOUNT_NAME in message \
                and self.names[message[ACCOUNT_NAME]] == client:
            response = RESPONSE_202
            response[DATA] = [user[0] for user in self.db.get_users()]
            try:
                send_message(client, response)
            except OSError:
                self.delete_client(client)
        elif ACTION in message and message[ACTION] == GET_PUBLIC_KEY and ACCOUNT_NAME in message:
            response = RESPONSE_511
            response[BIN] = self.db.get_public_key(message[ACCOUNT_NAME])
            if response[BIN]:
                try:
                    send_message(client, response)
                except OSError:
                    self.delete_client(client)
            else:
                response = RESPONSE_400
                response[ERROR] = 'Нет публичного ключа для данного пользователя'
                try:
                    send_message(client, response)
                except OSError:
                    self.delete_client(client)
        else:
            response = RESPONSE_400
            response[ERROR] = 'Запрос некорректен.'
            try:
                send_message(client, response)
            except OSError:
                self.delete_client(client)

    def autorize_user(self, message, transport):
        """
        :param message: словарь, содержащий данные об обрабатываемом сообщении
        :param transport: сокет клиента

        Метод, отвечающий за авторизацию пользователя на сервере. Проверяет не подключён ли уже пользователь к серверу и
        зарегистрирован ли на сервере, затем через алгоритм хэширования проверяет введённый пароль и осуществляет
        подключение пользователя, если всё в порядке
        """

        logger.debug(f'Start auth process for {message[USER]}')
        if message[USER][ACCOUNT_NAME] in self.names.keys():
            response = RESPONSE_400
            response[ERROR] = 'Имя пользователя уже занято.'
            try:
                logger.debug(f'Username busy, sending {response}')
                send_message(transport, response)
            except OSError:
                logger.debug('OS Error')
                pass
            self.clients.remove(transport)
            transport.close()
        elif not self.db.check_user(message[USER][ACCOUNT_NAME]):
            response = RESPONSE_400
            response[ERROR] = 'Пользователь не зарегистрирован.'
            try:
                logger.debug(f'Unknown username, sending {response}')
                send_message(transport, response)
            except OSError:
                pass
            self.clients.remove(transport)
            transport.close()
        else:
            logger.debug('Correct username, starting passwd check.')
            message_auth = RESPONSE_511
            random_str = binascii.hexlify(os.urandom(64))
            message_auth[BIN] = random_str.decode('ascii')
            hash = hmac.new(self.db.get_hash(message[USER][ACCOUNT_NAME]), random_str, 'MD5').digest()
            logger.debug(f'Auth message = {message_auth}')
            try:
                send_message(transport, message_auth)
                response = get_message(transport)
            except OSError as err:
                logger.debug('Error in auth, data:', exc_info=err)
                transport.close()
                return
            client_digest = binascii.a2b_base64(response[BIN])
            if RESPONSE in response and response[RESPONSE] == 511 and \
                    hmac.compare_digest(hash, client_digest):
                self.names[message[USER][ACCOUNT_NAME]] = transport
                client_ip, client_port = transport.getpeername()
                try:
                    send_message(transport, RESPONSE_200)
                except OSError:
                    self.delete_client(message[USER][ACCOUNT_NAME])
                self.db.user_login(
                    message[USER][ACCOUNT_NAME],
                    client_ip,
                    client_port,
                    message[USER][PUBLIC_KEY])
            else:
                response = RESPONSE_400
                response[ERROR] = 'Неверный пароль.'
                try:
                    send_message(transport, response)
                except OSError:
                    pass
                self.clients.remove(transport)
                transport.close()

    def service_update_lists(self):
        """
        Метод, осуществляющий обновление словаря пользователей сервера
        """

        for client in self.names:
            try:
                send_message(self.names[client], RESPONSE_205)
            except OSError:
                self.delete_client(self.names[client])




from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QLabel, QTableView
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import QTimer
from server_dist.server.server_files.stats_window import StatsWindow
from server_dist.server.server_files.config_window import ConfigWindow
from server_dist.server.server_files.add_user import AddUserDialog
from server_dist.server.server_files.delete_user import DeleteUserDialog


class MainWindow(QMainWindow):
    """
    Класс, определяющий и создающий основное окно графического интерфейса сервера
    """

    def __init__(self, db, server, config):
        super().__init__()

        self.db = db
        self.server = server
        self.config = config

        self.exitAction = QAction('Выход', self)
        self.exitAction.setShortcut('Ctrl+Q')
        self.exitAction.triggered.connect(qApp.quit)

        self.refresh_button = QAction('Обновить список', self)
        self.config_button = QAction('Настройки сервера', self)
        self.add_user_button = QAction('Регистрация пользователя', self)
        self.delete_user_button = QAction('Удаление пользователя', self)
        self.show_stats_button = QAction('История клиентов', self)

        self.statusBar()
        self.statusBar().showMessage('Server Working')
        self.toolbar = self.addToolBar('MainBar')
        self.toolbar.addAction(self.exitAction)
        self.toolbar.addAction(self.refresh_button)
        self.toolbar.addAction(self.show_stats_button)
        self.toolbar.addAction(self.config_button)
        self.toolbar.addAction(self.add_user_button)
        self.toolbar.addAction(self.delete_user_button)

        self.setFixedSize(800, 600)
        self.setWindowTitle('Messaging Server alpha release')

        self.label = QLabel('Список подключённых клиентов:', self)
        self.label.setFixedSize(240, 15)
        self.label.move(10, 25)

        self.active_clients_table = QTableView(self)
        self.active_clients_table.move(10, 45)
        self.active_clients_table.setFixedSize(780, 400)

        self.timer = QTimer()
        self.timer.timeout.connect(self.create_active_users_model)
        self.timer.start(1000)

        self.refresh_button.triggered.connect(self.create_active_users_model)
        self.show_stats_button.triggered.connect(self.show_stats)
        self.config_button.triggered.connect(self.server_config)
        self.add_user_button.triggered.connect(self.add_user)
        self.delete_user_button.triggered.connect(self.delete_user)

        self.show()

    def create_active_users_model(self):
        """
        Метод, заполняющий таблицу активных пользователей свежими данными из базы данных сервера
        """

        list_users = self.db.get_active_users()
        list = QStandardItemModel()
        list.setHorizontalHeaderLabels(
            ['Имя Клиента', 'IP Адрес', 'Порт', 'Время подключения'])
        for row in list_users:
            user, ip, port, time = row
            user = QStandardItem(user)
            user.setEditable(False)
            ip = QStandardItem(ip)
            ip.setEditable(False)
            port = QStandardItem(str(port))
            port.setEditable(False)
            time = QStandardItem(str(time.replace(microsecond=0)))
            time.setEditable(False)
            list.appendRow([user, ip, port, time])
        self.active_clients_table.setModel(list)
        self.active_clients_table.resizeColumnsToContents()
        self.active_clients_table.resizeRowsToContents()

    def show_stats(self):
        """
        Метод, вызывающий окно со статистикой пользователей
        """

        global stats_window
        stats_window = StatsWindow(self.db)
        stats_window.show()

    def server_config(self):
        """
        Метод, вызывающий окно изменения настроек сервера
        """

        global config_window
        config_window = ConfigWindow(self.config)

    def add_user(self):
        """
        Метод, вызывающий окно создания нового пользователя
        """

        global add_user_window
        add_user_window = AddUserDialog(self.db, self.server)
        add_user_window.show()

    def delete_user(self):
        """
        Метод, вызывающий окно удаления пользователя
        """

        global delete_user_window
        delete_user_window = DeleteUserDialog(self.db, self.server)
        delete_user_window.show()



import sys
import logging
import socket

# метод определения модуля, источника запуска.
if sys.argv[0].find('client_dist') == -1:
    #если не клиент то сервер!
    logger = logging.getLogger('server_dist')
else:
    # ну, раз не сервер, то клиент
    logger = logging.getLogger('client_dist')


def log(func_to_log):
    """
    :param func_to_log: функция для декорирования
    Функция-декоратор, осуществляющая логирование других функций
    """

    def log_saver(*args, **kwargs):
        logger.debug(f'Была вызвана функция {func_to_log.__name__} c параметрами {args} , {kwargs}. Вызов из модуля {func_to_log.__module__}')
        ret = func_to_log(*args, **kwargs)
        return ret
    return log_saver


def login_required(func):
    """
    :param func: функция для декорирования
    Функция-декоратор, проверяющая залогинен ли клиент на сервере
    """

    def checker(*args, **kwargs):
        from server_dist.server.server_files import MessageProcessor
        from server_dist.server.common.variables import ACTION, PRESENCE
        if isinstance(args[0], MessageProcessor):
            found = False
            for arg in args:
                if isinstance(arg, socket.socket):
                    for client in args[0].names:
                        if args[0].names[client] == arg:
                            found = True

            for arg in args:
                if isinstance(arg, dict):
                    if ACTION in arg and arg[ACTION] == PRESENCE:
                        found = True
            if not found:
                raise TypeError
        return func(*args, **kwargs)

    return checker


from sqlalchemy import *
from sqlalchemy.orm import mapper, sessionmaker
import datetime


class ServerDB:
    """
    Класс, определяющий, создающий и изменяющий серверную базу данных
    """

    class User:
        def __init__(self, name, hash):
            self.id = None
            self.name = name
            self.last_login = datetime.datetime.now()
            self.hash = hash
            self.public_key = None

        def __repr__(self):
            return f'<User: {self.id}-{self.name}>'

    class ActiveUser:
        def __init__(self, user, login_time, login_ip, login_port):
            self.id = None
            self.user = user
            self.login_time = login_time
            self.login_ip = login_ip
            self.login_port = login_port

        def __repr__(self):
            return f'<Active User: {self.id}-{self.user}>'

    class LoginHistory:
        def __init__(self, user, login_time, login_ip, login_port):
            self.id = None
            self.user = user
            self.login_time = login_time
            self.login_ip = login_ip
            self.login_port = login_port

        def __repr__(self):
            return f'<Login: {self.id}-{self.user}-{self.login_time}>'

    class Contact:
        def __init__(self, user, contact):
            self.id = None
            self.user = user
            self.contact = contact

        def __repr__(self):
            return f'<Contact: {self.id}-{self.user}-{self.contact}>'

    class MessageHistory:
        def __init__(self, user):
            self.user = user
            self.sent = 0
            self.received = 0

        def __repr__(self):
            return f'<MessageHistory: {self.user}-{self.sent}-{self.received}>'

    def __init__(self, path):
        self.engine = create_engine(f'sqlite:///{path}.db3', echo=False, pool_recycle=7200,
                                    connect_args={'check_same_thread': False})
        self.metadata = MetaData()

        users_table = Table('users', self.metadata,
                            Column('id', Integer, primary_key=True),
                            Column('name', String, unique=True),
                            Column('last_login', DateTime),
                            Column('hash', String),
                            Column('public_key', Text)
                            )

        active_users_table = Table('active_users', self.metadata,
                                   Column('id', Integer, primary_key=True),
                                   Column('user', ForeignKey('users.id'), unique=True),
                                   Column('login_time', DateTime),
                                   Column('login_ip', String),
                                   Column('login_port', String)
                                   )

        login_history_table = Table('login_history', self.metadata,
                                    Column('id', Integer, primary_key=True),
                                    Column('user', ForeignKey('users.id')),
                                    Column('login_time', DateTime),
                                    Column('login_ip', String),
                                    Column('login_port', String)
                                    )

        contacts_table = Table('contacts', self.metadata,
                               Column('id', Integer, primary_key=True),
                               Column('user', ForeignKey('users.id')),
                               Column('contact', ForeignKey('users.id'))
                               )

        message_history_table = Table('message_history', self.metadata,
                                      Column('user', ForeignKey('users.id'), primary_key=True),
                                      Column('sent', Integer),
                                      Column('received', Integer)
                                      )

        self.metadata.create_all(self.engine)

        mapper(self.User, users_table)
        mapper(self.ActiveUser, active_users_table)
        mapper(self.LoginHistory, login_history_table)
        mapper(self.Contact, contacts_table)
        mapper(self.MessageHistory, message_history_table)

        self.session = sessionmaker(bind=self.engine)()
        self.session.query(self.ActiveUser).delete()
        self.session.commit()

    def add_user(self, name, hash):
        """
        :param name: имя нового пользователя
        :param hash: хэш пароля нового пользователя

        Метод, добавляющий в таблицу пользователей нового пользователя
        """

        user = self.User(name, hash)
        self.session.add(user)
        self.session.commit()
        self.session.add(self.MessageHistory(user.id))
        self.session.commit()

    def delete_user(self, name):
        """
        :param name: имя пользователя, которого нужно удалить

        Метод, удаляющий пользователя из базы данных
        """

        user = self.session.query(self.User).filter_by(name=name).first()
        self.session.query(self.ActiveUser).filter_by(user=user.id).delete()
        self.session.query(self.LoginHistory).filter_by(user=user.id).delete()
        self.session.query(self.Contact).filter_by(user=user.id).delete()
        self.session.query(self.Contact).filter_by(contact=user.id).delete()
        self.session.query(self.MessageHistory).filter_by(user=user.id).delete()
        self.session.query(self.User).filter_by(id=user.id).delete()
        self.session.commit()

    def user_login(self, name, ip, port, key):
        """
        :param name: имя пользователя
        :param ip: адрес подключения
        :param port: порт подключения
        :param key:  публичный ключ пользователя

        Метод, сохраняющий данные о входе пользователя на сервер в базе данных сервера
        """

        query = self.session.query(self.User).filter_by(name=name)

        if query.count():
            user = query.first()
            user.last_login = datetime.datetime.now()
            if user.public_key != key:
                user.public_key = key
        else:
            raise ValueError('Пользователь не зарегистрирован')

        self.session.add(self.ActiveUser(user.id, datetime.datetime.now(), ip, port))
        self.session.add(self.LoginHistory(user.id, datetime.datetime.now(), ip, port))

        self.session.commit()

    def user_logout(self, name):
        """
        :param name: имя пользователя

        Метод, удаляющий пользователя из таблицы активных пользователей
        """

        self.session.query(self.ActiveUser).filter_by(user=self.session.query(self.User)
                                                      .filter_by(name=name).first().id).delete()
        self.session.commit()

    def get_hash(self, name):
        """
        :param name: имя пользователя

        Метод, возвращающий хэш пароля пользователя из базы данных
        """

        return self.session.query(self.User).filter_by(name=name).first().hash

    def get_public_key(self, name):
        """
        :param name: имя пользователя

        Метод, возвращающий публичный ключ пользователя из базы данных
        """

        return self.session.query(self.User).filter_by(name=name).first().public_key

    def get_users(self):
        """
        Метод, возвращающий список доступных пользователей из базы данных сервера
        """

        return self.session.query(self.User.name, self.User.last_login).all()

    def check_user(self, name):
        """
        :param name: имя пользователя

        Метод, проверяющий зарегистрирован ли пользователь на сервере
        """

        if self.session.query(self.User).filter_by(name=name).count():
            return True
        else:
            return False

    def get_active_users(self):
        """
        Метод, возвращающий список активных пользователей из базы данных
        """

        return self.session.query(self.User.name, self.ActiveUser.login_ip, self.ActiveUser.login_port, self.ActiveUser.login_time).join(self.User).all()

    def get_login_history(self, name):
        """
        :param name: имя пользователя

        Метод, возвращающий истроию входа пользователя на сервер из базы данных
        """

        return self.session.query(self.LoginHistory).filter_by(user=self.session.query(self.User)
                                                               .filter_by(name=name).first().id).all()

    def get_contacts(self, name):
        """
        :param name: имя пользователя

        Метод, возвращающий список контактов пользователя из базы данных
        """

        return [contact[1] for contact in self.session.query(self.Contact, self.User.name).filter_by(user=self.session.query(self.User)
                                                          .filter_by(name=name).first().id).join(self.User, self.Contact.contact == self.User.id).all()]

    def get_message_history(self, name=None):
        """
        :param name: имя пользователя

        Метод возвращающий историю сообщений конкретного пользователя или всех пользователей
        """

        if not name:
            return self.session.query(self.User.name, self.User.last_login, self.MessageHistory.sent, self.MessageHistory.received).join(self.User).all()
        return self.session.query(self.User.name, self.User.last_login, self.MessageHistory.sent, self.MessageHistory.received)\
            .filter_by(user=self.session.query(self.User).filter_by(name=name).first().id).join(self.User).all()

    def add_contact(self, user_name, contact_name):
        """
        :param user_name: имя пользователя
        :param contact_name: имя пользователя, которого нужно добавить в контакты

        Метод, добавляющий одного пользователя в список контактов другого пользователя
        """

        user = self.session.query(self.User).filter_by(name=user_name).first().id
        contact = self.session.query(self.User).filter_by(name=contact_name).first().id
        self.session.add(self.Contact(user, contact))
        self.session.commit()

    def delete_contact(self, user_name, contact_name):
        """
        :param user_name: имя пользователя
        :param contact_name: имя пользователя, которого нужно удалить из контактов

        Метод, удаляющий одного пользователя из списка контактов другого пользователя
        """

        user = self.session.query(self.User).filter_by(name=user_name).first().id
        contact = self.session.query(self.User).filter_by(name=contact_name).first().id
        self.session.query(self.Contact).filter_by(user=user, contact=contact).delete()
        self.session.commit()

    def process_message(self, sender_name, receiver_name):
        """
        :param sender_name: имя отправителя
        :param receiver_name: имя получателя

        Метод, сохраняющий сообщение в таблице истории сообщений
        """

        sender = self.session.query(self.User).filter_by(name=sender_name).first().id
        receiver = self.session.query(self.User).filter_by(name=receiver_name).first().id

        self.session.query(self.MessageHistory).filter_by(user=sender).first().sent += 1
        self.session.query(self.MessageHistory).filter_by(user=receiver).first().received += 1

        self.session.commit()


if __name__ == '__main__':
    test_db = ServerDB()
    test_db.user_login('client_1', '192.168.1.4', 8080)
    test_db.user_login('client_2', '192.168.1.5', 7777)

    print(' ---- test_db.get_active_users() ----')
    print(test_db.get_active_users())

    test_db.user_logout('client_1')
    print(' ---- test_db.get_active_users() after logout client_1 ----')
    print(test_db.get_active_users())

    print(' ---- test_db.login_history(client_1) ----')
    print(test_db.get_login_history('client_1'))

    print(' ---- test_db.get_users() ----')
    print(test_db.get_users())

    print(' ---- test_db.get_contacts() of client_1----')
    print(test_db.get_contacts('client_1'))

    print(' ---- test_db.add_contact() client_2 to client_1----')
    test_db.add_contact('client_1', 'client_2')
    print(test_db.get_contacts('client_1'))

    print(' ---- test_db.delete_contact() client_2 of client_1----')
    test_db.delete_contact('client_1', 'client_2')
    print(test_db.get_contacts('client_1'))

    print(' ---- test_db.get_message_history() of client_1 and client_2----')
    print(test_db.get_message_history('client_1'))
    print(test_db.get_message_history('client_2'))

    print(' ---- test_db.count_message() client_1 to client_2----')
    test_db.process_message('client_1', 'client_2')
    print(test_db.get_message_history('client_1'))
    print(test_db.get_message_history('client_2'))

    test_db.user_logout('client_2')
    test_db.session.query(test_db.User).filter(test_db.User.name.in_(['client_1', 'client_2'])).delete()

    print(' ---- test_db.users_list() ----')
    print(test_db.get_users())



from setuptools import setup, find_packages

setup(name="barrakuda_messenger_server",
      version="0.1.0",
      description="barrakuda_messenger_server",
      author="barrakuda8",
      author_email="brpslelush@gmail.com",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex']
      )



import configparser
import os.path
import argparse
from common.utils import *
from common.decos import log
from server_dist.server.server_files import ServerDB
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import Qt
from server_dist.server.server_files.main_window import MainWindow
from server_dist.server.server_files import MessageProcessor
import logging

logger = logging.getLogger('server_dist')


@log
def arg_parser(default_port, default_address):
    """
    Метод, разбирающий команду и вытаскивающий из неё параметры
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', default=default_port, type=int, nargs='?')
    parser.add_argument('-a', default=default_address, nargs='?')
    parser.add_argument('--no_gui', action='store_true')

    namespace = parser.parse_args(sys.argv[1:])
    listen_address = namespace.a
    listen_port = namespace.p
    gui_flag = namespace.no_gui

    return listen_address, listen_port, gui_flag


@log
def config_load():
    config = configparser.ConfigParser()
    dir_path = os.path.dirname(os.path.realpath(__file__))
    config.read(f"{dir_path}/{'server_dist+++.ini'}")
    if 'SETTINGS' in config:
        return config
    else:
        config.add_section('SETTINGS')
        config.set('SETTINGS', 'Default_port', str(DEFAULT_PORT))
        config.set('SETTINGS', 'Listen_Address', '')
        config.set('SETTINGS', 'Database_path', '')
        config.set('SETTINGS', 'Database_file', 'server_db')
        return config


def main():
    config = config_load()
    listen_address, listen_port, gui_flag = arg_parser(config['SETTINGS']['Default_port'], config['SETTINGS']['Listen_address'])
    db = ServerDB(os.path.join(config['SETTINGS']['Database_path'], config['SETTINGS']['Database_file']))

    server = MessageProcessor(listen_address, listen_port, db)
    server.start()

    if gui_flag:
        while True:
            command = input('Введите exit для завершения работы сервера.')
            if command == 'exit':
                server.running = False
                server.join()
                break

    else:
        server_app = QApplication(sys.argv)
        server_app.setAttribute(Qt.AA_DisableWindowContextHelpButton)
        main_window = MainWindow(db, server, config)
        server_app.exec_()
        server.running = False


if __name__ == '__main__':
    main()
